/*  Basic m68hc05 simulator
    Copyright 2004 Free Software Foundation, Inc.
    Copyright 2004 Martin Kgler
    Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
    
    */


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

typedef enum
{
  M68HC05_NONE,
  M68HC05_IMM,
  M68HC05_OFFSET8,
  M68HC05_OFFSET16,
  M68HC05_INDEX0,
  M68HC05_INDEX1,
  M68HC05_INDEX2,
  M68HC05_PCREL,
  M68HC05_OFFSET8_PCREL,
} M68HC05_PARAM;

struct m68hc05_opcodes_s
{
  char *name;
  M68HC05_PARAM par;
  unsigned char opcode;
};

#define M68HC05_INS(NAME, PAR, OPCODE) \
{NAME, PAR, OPCODE},

const struct m68hc05_opcodes_s m68hc05_opcodes[] = {
#include "opcode/m68hc05.h"
  {NULL, 0, 0}
};

char opcode_done[256] = { 0 };

int
main ()
{
  const struct m68hc05_opcodes_s *op = m68hc05_opcodes;
  int cycles = 0;

  printf ("/* File generated by gencode */\n");
  printf ("#include \"sim-main.h\"\n\n");

  printf ("void cpu_interp_m68hc05(sim_cpu *proc)\n");
  printf ("{\n");
  printf ("uint16 addr,tval;uint8 op,op1,op2;\n\n");
  printf ("op=cpu_fetch8 (proc);\n");
  printf ("switch(op)\n{\n");

  while (op->name)
    {
      if (opcode_done[op->opcode])
	{
	  op++;
	  continue;
	}
      cycles = 2;
      opcode_done[op->opcode] = 1;
      printf ("case 0x%02X: /*%s*/\n", op->opcode, op->name);
      switch (op->par)
	{
	case M68HC05_NONE:
	  break;
	case M68HC05_IMM:
	  printf ("op1=cpu_fetch8(proc);\n");
	  break;
	case M68HC05_OFFSET8:
	  printf ("addr=cpu_fetch8(proc);\n");
	  printf ("op1=memory_read8(proc,addr);\n");
	  cycles += 1;
	  break;
	case M68HC05_OFFSET16:
	  printf ("addr=cpu_fetch16(proc);\n");
	  printf ("op1=memory_read8(proc,addr);\n");
	  cycles += 2;
	  break;
	case M68HC05_INDEX0:
	  printf ("addr=cpu_get_x(proc);\n");
	  printf ("op1=memory_read8(proc,addr);\n");
	  cycles += 1;
	  break;
	case M68HC05_INDEX1:
	  printf ("addr=cpu_fetch8(proc)+cpu_get_x(proc);\n");
	  printf ("op1=memory_read8(proc,addr);\n");
	  cycles += 2;
	  break;
	case M68HC05_INDEX2:
	  printf ("addr=cpu_fetch16(proc)+cpu_get_x(proc);\n");
	  printf ("op1=memory_read8(proc,addr);\n");
	  cycles += 3;
	  break;
	case M68HC05_PCREL:
	  printf
	    ("addr=((signed char)(cpu_fetch8(proc)))+cpu_get_pc(proc);\n");
	  cycles += 1;
	  break;
	case M68HC05_OFFSET8_PCREL:
	  printf ("op1=memory_read8(proc,cpu_fetch8(proc));\n");
	  printf
	    ("addr=((signed char)(cpu_fetch8(proc)))+cpu_get_pc(proc);\n");
	  cycles += 3;
	  break;
	}
      if (!strcmp (op->name, "txa"))
	printf ("cpu_set_a(proc,cpu_get_x(proc));\n");
      else if (!strcmp (op->name, "tst"))
	{
	  cycles += 1;
	  printf ("cpu_ccr_update_NZ(proc,memory_read8(proc,addr));\n");
	}
      else if (!strcmp (op->name, "tsta"))
	{
	  cycles += 1;
	  printf ("cpu_ccr_update_NZ(proc,cpu_get_a(proc));\n");
	}
      else if (!strcmp (op->name, "tstx"))
	{
	  cycles += 1;
	  printf ("cpu_ccr_update_NZ(proc,cpu_get_x(proc));\n");
	}
      else if (!strcmp (op->name, "tax"))
	printf ("cpu_set_x(proc,cpu_get_a(proc));\n");
      else if (!strcmp (op->name, "sub"))
	{
	  printf ("tval=cpu_get_a(proc)-op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "stx"))
	{
	  cycles++;
	  printf ("memory_write8(proc,addr,cpu_get_x(proc));\n");
	  printf ("cpu_ccr_update_NZ(proc,cpu_get_x(proc));\n");
	}
      else if (!strcmp (op->name, "sta"))
	{
	  cycles++;
	  printf ("memory_write8(proc,addr,cpu_get_a(proc));\n");
	  printf ("cpu_ccr_update_NZ(proc,cpu_get_a(proc));\n");
	}
      else if (!strcmp (op->name, "sei"))
	printf ("cpu_set_ccr_I(proc,1);\n");
      else if (!strcmp (op->name, "sec"))
	printf ("cpu_set_ccr_C(proc,1);\n");
      else if (!strcmp (op->name, "sbc"))
	{
	  printf ("tval=cpu_get_a(proc)-op1-cpu_get_ccr_C(proc);\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "rts"))
	{
	  cycles += 4;
	  printf ("cpu_set_pc(proc,cpu_pop_uint16(proc));\n");
	}
      else if (!strcmp (op->name, "rsp"))
	printf ("cpu_set_sp(proc,0xff);\n");
      else if (!strcmp (op->name, "ror"))
	{
	  cycles += 2;
	  printf ("tval=(op1|cpu_get_ccr_C(proc)<<8);\n");
	  printf ("memory_write8(proc,addr,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "rora"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc)|cpu_get_ccr_C(proc)<<8);\n");
	  printf ("cpu_set_a(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "rorx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc)|cpu_get_ccr_C(proc)<<8);\n");
	  printf ("cpu_set_x(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "rol"))
	{
	  cycles += 2;
	  printf ("tval=(op1<<1)|cpu_get_ccr_C(proc);\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "rola"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc)<<1)|cpu_get_ccr_C(proc);\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "rolx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc)<<1)|cpu_get_ccr_C(proc);\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "ora"))
	{
	  printf ("tval=cpu_get_a(proc)|op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "nop"))
	printf ("\n");
      else if (!strcmp (op->name, "neg"))
	{
	  cycles += 2;
	  printf ("tval=-op1;\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "nega"))
	{
	  cycles += 1;
	  printf ("tval=-cpu_get_a(proc);\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "negx"))
	{
	  cycles += 1;
	  printf ("tval=-cpu_get_x(proc);\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "mul"))
	{
	  cycles += 9;
	  printf ("tval=cpu_get_x(proc)*cpu_get_a(proc);\n");
	  printf
	    ("cpu_set_x(proc,(tval>>8)&0xff);cpu_set_a(proc,tval&0xff);\n");
	  printf ("cpu_set_ccr_H(proc,0);cpu_set_ccr_C(proc,0);\n");
	}
      else if (!strcmp (op->name, "lsr"))
	{
	  cycles += 2;
	  printf ("tval=(op1);\n");
	  printf ("memory_write8(proc,addr,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "lsra"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc));\n");
	  printf ("cpu_set_a(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "lsrx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc));\n");
	  printf ("cpu_set_x(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "ldx"))
	{
	  printf ("tval=op1;\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "lda"))
	{
	  printf ("tval=op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "jsr"))
	{
	  printf ("cpu_push_uint16(proc,cpu_get_pc(proc));\n");
	  printf ("cpu_set_pc(proc,addr);\n");
	  cycles += 2;
	}
      else if (!strcmp (op->name, "jmp"))
	{
	  printf ("cpu_set_pc(proc,addr);\n");
	  cycles -= 1;
	}
      else if (!strcmp (op->name, "inc"))
	{
	  cycles += 2;
	  printf ("tval=(op1)+1;\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "inca"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc))+1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "incx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc))+1;\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "eor"))
	{
	  printf ("tval=cpu_get_a(proc)^op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "dec"))
	{
	  cycles += 2;
	  printf ("tval=(op1)+1;\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "deca"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc))+1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "decx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc))+1;\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "cpx"))
	{
	  printf ("tval=cpu_get_x(proc)-op1;\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "com"))
	{
	  cycles += 2;
	  printf ("tval=0xff-(op1);\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	  printf ("cpu_set_ccr_C(proc,1);\n");
	}
      else if (!strcmp (op->name, "coma"))
	{
	  cycles += 1;
	  printf ("tval=0xff-(cpu_get_a(proc));\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	  printf ("cpu_set_ccr_C(proc,1);\n");
	}
      else if (!strcmp (op->name, "comx"))
	{
	  cycles += 1;
	  printf ("tval=0xff-(cpu_get_x(proc));\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	  printf ("cpu_set_ccr_C(proc,1);\n");
	}
      else if (!strcmp (op->name, "cmp"))
	{
	  printf ("tval=cpu_get_a(proc)-op1;\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "clr"))
	{
	  cycles += 2;
	  printf ("tval=0;\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "clra"))
	{
	  cycles += 1;
	  printf ("tval=0;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "clrx"))
	{
	  cycles += 1;
	  printf ("tval=0;\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "cli"))
	printf ("cpu_set_ccr_I(proc,0);\n");
      else if (!strcmp (op->name, "clc"))
	printf ("cpu_set_ccr_C(proc,0);\n");
      else if (!strcmp (op->name, "bsr"))
	{
	  printf ("cpu_push_uint16(proc,cpu_get_pc(proc));\n");
	  printf ("cpu_set_pc(proc,addr);\n");
	  cycles += 3;
	}
      else if (!strncmp (op->name, "bset.", 5))
	{
	  cycles += 2;
	  printf ("memory_write8(proc,addr,op1|(1<<%d));\n",
		  op->name[5] - '0');
	}
      else if (!strncmp (op->name, "brset.", 6))
	{
	  cycles += 2;
	  printf ("if((op1&(1<<%d))!=0)cpu_set_pc(proc,addr);\n",
		  op->name[6] - '0');
	}
      else if (!strcmp (op->name, "brn"))
	cycles++;
      else if (!strncmp (op->name, "brclr.", 6))
	{
	  cycles += 2;
	  printf ("if((op1&(1<<%d))==0)cpu_set_pc(proc,addr);\n",
		  op->name[6] - '0');
	}
      else if (!strcmp (op->name, "bra"))
	{
	  cycles++;
	  printf ("if(1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bpl"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_N(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bne"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_Z(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bms"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_I(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bmi"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_N(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bmc"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_I(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bls"))
	{
	  cycles++;
	  printf
	    ("if(cpu_get_ccr_C(proc)==1||cpu_get_ccr_Z(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bit"))
	{
	  printf ("tval=cpu_get_a(proc)^op1;\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "bhs"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_C(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bhi"))
	{
	  cycles++;
	  printf
	    ("if(cpu_get_ccr_C(proc)==0&&cpu_get_ccr_Z(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bhcs"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_H(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bhcc"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_H(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "beq"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_Z(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "bcs"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_C(proc)==1)cpu_set_pc(proc,addr);\n");
	}
      else if (!strncmp (op->name, "bclr.", 5))
	{
	  cycles += 2;
	  printf ("memory_write8(proc,addr,op1&~(1<<%d));\n",
		  op->name[5] - '0');
	}
      else if (!strcmp (op->name, "bcc"))
	{
	  cycles++;
	  printf ("if(cpu_get_ccr_C(proc)==0)cpu_set_pc(proc,addr);\n");
	}
      else if (!strcmp (op->name, "asr"))
	{
	  cycles += 2;
	  printf ("tval=(op1)|((op1&0x80)<<1);\n");
	  printf ("memory_write8(proc,addr,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "asra"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc))|((cpu_get_a(proc)&0x80)<<1);\n");
	  printf ("cpu_set_a(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "asrx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc))|((cpu_get_x(proc)&0x80)<<1);\n");
	  printf ("cpu_set_x(proc,tval>>1);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval>>1);\n");
	  printf ("cpu_set_ccr_C(proc,tval&0x1);\n");
	}
      else if (!strcmp (op->name, "asl"))
	{
	  cycles += 2;
	  printf ("tval=(op1<<1);\n");
	  printf ("memory_write8(proc,addr,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "asla"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_a(proc)<<1);\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "aslx"))
	{
	  cycles += 1;
	  printf ("tval=(cpu_get_x(proc)<<1);\n");
	  printf ("cpu_set_x(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	}
      else if (!strcmp (op->name, "and"))
	{
	  printf ("tval=cpu_get_a(proc)&op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZ(proc,tval);\n");
	}
      else if (!strcmp (op->name, "add"))
	{
	  printf ("tval=cpu_get_a(proc)+op1;\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	  printf
	    ("cpu_set_ccr_H(proc,((cpu_get_a(proc)&0x0f)+(op1&0x0f))&0x10);\n");
	}
      else if (!strcmp (op->name, "adc"))
	{
	  printf ("tval=cpu_get_a(proc)+op1+cpu_get_ccr_C(proc);\n");
	  printf ("cpu_set_a(proc,tval);\n");
	  printf ("cpu_ccr_update_NZC(proc,tval);\n");
	  printf
	    ("cpu_set_ccr_H(proc,((cpu_get_a(proc)&0x0f)+(op1&0x0f)+cpu_get_ccr_C(proc))&0x10);\n");
	}
      else
	printf ("goto error;\n");
      printf ("cpu_add_cycles(proc,%d);\n", cycles);
      printf ("break;\n");
      op++;
    }
  printf ("default:\n");
  printf ("error:\n");
  printf ("cpu_special (proc,op);\n");
  printf ("}\n");
  printf ("}\n");

  return 0;
}
